from collections import Counter
from re import finditer, compile, match
import textwrap
from typing import List
import inflect
import os
import boto3
from config import get_logger, YAML_IGNORE_PREFIX
import botocore

# from functools import lru_cache

inflect_engine = inflect.engine()  # used for singular/plural word comparing
logger = get_logger(__name__)

_camel_case_regex_compiled = compile(
    r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)
_terraform_aws_resource_type_pattern_compiled = compile(r"^aws_[a-zA-Z0-9_-]+$")


def is_terraform_aws_resource_type(resource_type: str) -> bool:
    """Check for terraform aws resource type being valid

    Args:
        resource_type (str): Check for terraform aws resource type being valid

    Returns:
        bool: is resource type valid
    """
    return bool(match(_terraform_aws_resource_type_pattern_compiled, resource_type))


def find_all_yaml_files(directory: str) -> List[str]:
    """Find all .yaml files in a directory with the exception of files starting with "_".

    Args:
        directory (str): Directory to search for yaml files

    Returns:
        List[str]: List of yaml file paths
    """
    yaml_files = []
    for root, _, files in os.walk(directory):
        for file in files:
            # support any yaml file with the exception of files starting with "YAML_IGNORE_PREFIX"
            if file.endswith(".yaml") and not file.startswith(YAML_IGNORE_PREFIX):
                yaml_files.append(os.path.join(root, file))
    return yaml_files


def _create_boto_session():
    """Tries to create a boto3 session with the given environment variables.
    Handles the exceptions and exits the program if it fails.

    Returns:
        boto3.session.Session: created boto3 session
    """
    profile_name = os.environ.get("AWS_PROFILE", False)
    region_name = os.environ.get("AWS_REGION", False) or os.environ.get(
        "AWS_DEFAULT_REGION", False
    )
    _kwargs_dict = {}
    if profile_name:
        _kwargs_dict["profile_name"] = profile_name
    if region_name:
        _kwargs_dict["region_name"] = region_name
    session = None

    try:
        if _kwargs_dict:
            session = boto3.session.Session(**_kwargs_dict)
        else:
            session = boto3.session.Session()
        return session
    except botocore.exceptions.BotoCoreError as e:
        logger.error(
            f"Failed to create boto3 session. Check [bold yellow]AWS_PROFILE, AWS_REGION or "
            f"AWS_DEFAULT_REGION[/] environment variables.\nException: [red bold]{str(e)}[/]"
        )
        exit(1)
    except Exception as e:
        logger.error(
            f"Unexpected error while creating boto3 session.\nException: [red bold]{str(e)}[/]"
        )
        exit(1)


def are_two_lists_same(list_one: List, list_two: List) -> bool:
    """Compares the contents of two lists"""
    return Counter(list_one) == Counter(list_two)


def is_word_in_a_list_of_words(word: str, list_of_words: List[str]) -> bool:
    """Checks if a word is a is in a list_of_words, case insensitive"""
    lower_word = word.lower()
    for a_word in list_of_words:
        if lower_word == a_word.lower():
            return True
    return False

    return word.lower() in []


def inform_about_develeoping_custom_resource_nodes():
    logger.debug(
        textwrap.dedent(
            """
        [bold yellow][FAILURE][/] It seems balcony [red bold]failed[/] to read this operation. 
        
        You can create Custom ResourceNodes to fix the failed operation. 
        Visit [bold][link=https://oguzhan-yilmaz.github.io/balcony/development/developing-custom-resource-nodes/]Custom Resource Development Documentation[/link][/] to learn more.
        
        You can also track the [bold][link=https://github.com/oguzhan-yilmaz/balcony/issues]Github Issues[/link][/] or create a new issue.
        """.lstrip(
                "\n"
            )
        )
    )


def ifind_similar_names_in_list(
    search_for: str, search_in_list: List[str]
) -> List[str]:
    """Case insensitive find in list"""
    result = []
    if not search_for or not search_in_list:
        return result
    lower_search_for = search_for.lower()

    for l_name in search_in_list:
        if l_name.lower().startswith(lower_search_for):
            result.append(l_name)
    return result


def str_relations(relations: List[dict]) -> str:
    """Stringify list of relations"""
    return ", ".join(
        [f"[{r.service_name}.{r.operation_name} > {r.target_path}]" for r in relations]
    )


def get_all_available_services(session: boto3.session.Session) -> List[str]:
    """Gets available services from boto3 session"""
    return session.get_available_services()


def compare_nouns(word1: str, word2: str) -> bool:
    """Singular/plural insensitive word comparison"""
    return inflect_engine.compare_nouns(word1, word2)


def compare_two_token_lists(
    token_list_one: List[str], token_list_two: List[str]
) -> bool:
    """Compares to word lists one by one in lowercase"""
    if len(token_list_one) != len(token_list_two):
        return False
    token_by_token_match = [
        w1.lower() == w2.lower() for w1, w2 in zip(token_list_one, token_list_two)
    ]
    has_token_by_token_match = all(token_by_token_match)
    return has_token_by_token_match


def icompare_two_token_lists(
    token_list_one: List[str], token_list_two: List[str]
) -> bool:
    """Singular/plural insensitive token list comparison"""
    if len(token_list_one) != len(token_list_two):
        return False
    token_by_token_match = [
        bool(inflect_engine.compare_nouns(w1.lower(), w2.lower()))
        for w1, w2 in zip(token_list_one, token_list_two)
    ]
    has_token_by_token_match = all(token_by_token_match)
    return has_token_by_token_match


def camel_case_split(identifier: str) -> List[str]:
    """Splits CamelCase string to it's tokens"""
    matches = finditer(_camel_case_regex_compiled, identifier)
    return [m.group(0) for m in matches]


def compare_two_camel_case_words(word1: str, word2: str) -> bool:
    token_list_1 = [_.lower() for _ in camel_case_split(word1)]
    token_list_2 = [_.lower() for _ in camel_case_split(word2)]
    return compare_two_token_lists(token_list_1, token_list_2)


def icompare_two_camel_case_words(word1: str, word2: str) -> bool:
    token_list_1 = [_.lower() for _ in camel_case_split(word1)]
    token_list_2 = [_.lower() for _ in camel_case_split(word2)]
    return icompare_two_token_lists(token_list_1, token_list_2)



from config import get_logger, get_rich_console
from errors import Error
from utils import inform_about_develeoping_custom_resource_nodes
import fnmatch  # unix like pattern matching
from collections.abc import Iterable
from typing import List, Dict, Tuple, Optional, Union
from botocore.exceptions import ClientError


logger = get_logger(__name__)
console = get_rich_console()


class ServiceReader:
    """
    Upon initialization ServiceReaders defines a dictionary called `response_data`.
    Caled operations are stored in this dictionary.
    ```json title="self.response_data data hierarchy"
    {
        "ResourceNodeName": {
            "OperationName1": [{response1 dict}, {response2 dict}, ...],
            "OperationName2": [{response1 dict},...],
        }
    }
    ```
    """

    def __init__(self, service_node: "ServiceNode") -> None:
        """Initializes the reader with the `service_node`.

        Args:
            service_node (ServiceNode): Associated ServiceNode.
        """
        self.service_node = service_node
        self.response_data = {}

    def call_operation(
        self, resource_node: "ResourceNode", operation_name: str, api_parameter: Dict, follow_pagination : Optional[bool] = False # noqa  
    ) -> Union[dict, bool]:
        """Calls the given AWS operation with `api_parameter` dict.
        Saves the response data on `self.response_data` and returns it.

        Args:
            resource_node: Operations Resource Node
            operation_name (str): Name of the operation
            api_parameter (dict): dictionary to call the operation with
            follow_pagination (Optional[bool]): If the operations output is truncated follow the pagination tokens.

        Returns:
            Union[dict, bool]: `False` or response got from AWS API
        """
        client = self.service_node.client

        if not resource_node:
            return False

        try:
            logger.debug(
                f"Calling operation: [bold blue]{operation_name}[/] with api parameters: {api_parameter}"
            )
            response = client._make_api_call(operation_name, api_parameter)
            # check if the operation is paginated
            pagination_tokens = (
                resource_node.get_pagination_token_output_to_parameter_name_mapping(
                    operation_name
                )
            )
            if follow_pagination and pagination_tokens:
                # try to find the name of the operation you'd want to fill out
                # and where to find it from in the response
                pagination_parameter_name = pagination_tokens.get("parameter_name")
                pagination_output_key = pagination_tokens.get("output_key")
                # we may find out the operation is paginated, but the queried resource might not be much in quantity
                # so, response might not have a Pagination token, although the operation is paginated
                page_value_in_response = response.get(pagination_output_key, False)
                if page_value_in_response:
                    paginated_api_parameters = api_parameter.copy()
                    paginated_api_parameters[
                        pagination_parameter_name
                    ] = page_value_in_response
                    self.call_operation(
                        resource_node, operation_name, paginated_api_parameters, follow_pagination=follow_pagination
                    )
        except ClientError as e:
            logger.debug(
                f"[red bold]FAILED: Calling Operation[/]. {operation_name}({api_parameter}). Exception: {str(e)}"
            )
            return False

        # removing ResponseMetadata, it is not needed
        response.pop("ResponseMetadata")
        response["__args__"] = api_parameter
        self.add_to_node_data(resource_node.name, operation_name, response)
        return response

    def search_operation_data(
        self, resource_node_name: str, operation_name: str
    ) -> Union[List[dict], bool]:
        """Get the currently read and available operation data.

        Args:
            resource_node_name (str): Name of the ResourceNode
            operation_name (str): Name of the Operation

        Returns:
            Union[List[dict], bool]: `False` or Operations data
        """
        resource_node_exists = (
            self.response_data.get(resource_node_name, False) != False  # noqa
        )
        if not resource_node_exists:
            return False
        resource_data: Dict = self.search_resource_node_data(resource_node_name)  # type: ignore
        result = resource_data.get(operation_name, [])
        return result

    def search_resource_node_data(self, resource_node_name: str) -> Union[dict, bool]:
        """Gets all data available for the ResourceNode, including all of its operations.

        Args:
            resource_node_name (str): Name of the ResourceNode

        Returns:
            Union[dict, bool]: `False` or all operations data of the given ResourceNode
        """
        resource_node_data = self.response_data.get(resource_node_name, False)
        return resource_node_data

    def clear_operations_data(
        self, resource_node_name: str, operation_name: str
    ) -> None:
        """Refreshes the operations data to empty list.

        Args:
            resource_node_name (str): Name of the ResourceNode
            operation_name (str): Name of the Operation
        """
        resource_node_exists = (
            self.response_data.get(resource_node_name, False) != False  # noqa
        )
        if resource_node_exists:
            self.response_data[resource_node_name][operation_name] = []

    def add_to_node_data(
        self, resource_node_name: str, operation_name: str, response: dict
    ) -> None:
        """Adds boto3 api response to operations existing data.

        Args:
            resource_node_name (str): Name of the ResourceNode
            operation_name (str): Name of the Operation
            response (dict): boto API response dict
        """
        resource_node_exists = (
            self.response_data.get(resource_node_name, False) != False # noqa
        )

        if not resource_node_exists:
            self.response_data[resource_node_name] = {}

        if operation_name not in self.response_data[resource_node_name]:
            self.response_data[resource_node_name][operation_name] = [response]
        else:
            self.response_data[resource_node_name][operation_name].append(response)

    def read_operation(
        self,
        resource_node_name: str,
        operation_name: str,
        match_patterns: Optional[List[str]] = None,
        refresh: Optional[bool] = False,
        follow_pagination: Optional[bool] = False
    ) -> Tuple[Union[List, bool], Union[Error, None]]:
        """Reads the given operation.
        If the operation is called with generated parameters, `match_patterns` can be used to filter the generated parameters.

        Args:
            resource_node_name (str): Name of the Resource Node
            operation_name (str): Name of the Operation
            match_patterns (List[str], optional): UNIX style patterns to filter matching generated parameters. Defaults to None.
            refresh (bool, optional): Get the cached data or force re-reading the operation. Defaults to False.
            follow_pagination (bool, optional): Follow pagination tokens. If not only set True, one page call will be made.

        Returns:
            Tuple[Union[List, bool], Union[Error, None]]: _description_
        """
        # if it has been read called already, return it if refresh is not set
        operation_markup = (
            f"[bold][green]{self.service_node.name}[/].[blue]{operation_name}[/][/]"
        )
        logger.debug(f"[underline][bold]Reading[/] {operation_markup}[/]")

        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            logger.debug(
                f"Failed to find the Resource Node while reading the {operation_markup}."
            )
            return None
        if refresh == True:  # noqa
            # remove the operation data from the response data,
            self.clear_operations_data(resource_node_name, operation_name)
        already_existing_data = self.search_operation_data(
            resource_node.name, operation_name
        )

        if already_existing_data and refresh == False:  # noqa
            logger.debug(
                f"[green]{resource_node.name}[/].[blue]{operation_name}[/] is already read. Returning already available data."
            )
            return already_existing_data

        if not resource_node:
            logger.debug(
                f"Failed to find the Resource Node while reading the {operation_markup}."
            )
            return False

        # try to automatically find the relations for this operation
        (
            relations_of_operation,
            relations_error,
        ) = resource_node.get_operations_relations(operation_name)
        success_finding_relations = relations_error is None
        required_parameters = (
            resource_node.get_required_parameter_names_from_operation_name(
                operation_name
            )
        )
        req_param_markup = f"[bold magenta]{', '.join(required_parameters)}[/]"

        def api_parameters_match_pattern(
            api_parameters: List[dict], patterns: List[str]
        ) -> List[dict]:
            if not patterns:
                return api_parameters
            matched_parameters = []

            for api_param in api_parameters:
                has_api_param_added = False
                for api_param_value in api_param.values():
                    for pattern in patterns:
                        if (
                            not has_api_param_added
                            and fnmatch.fnmatch(api_param_value, pattern)
                            and api_param not in matched_parameters
                        ):
                            matched_parameters.append(api_param)
                            has_api_param_added = True
            return matched_parameters

        if not success_finding_relations:
            logger.debug(f"[red]Error: {relations_error}: {operation_markup}")
            logger.debug(
                f"Failed to find the Relations for operation, required parameters: {req_param_markup}"
            )
            inform_about_develeoping_custom_resource_nodes()
            return False, relations_error

        if (
            relations_of_operation == True  # noqa
        ):  # True means no required parameters, so no relations
            # no relations means there are no related operations
            (
                generated_api_parameters,
                generation_error,
            ) = resource_node.generate_api_parameters_from_operation_data(
                operation_name, [], {}
            )

            if generation_error is not None:
                logger.debug(
                    f"Failed to generate api parameters for {operation_markup}. Error: {generation_error}"
                )
                inform_about_develeoping_custom_resource_nodes()
                return False, generation_error
            if isinstance(generated_api_parameters, Iterable):
                logger.debug(
                    f"Successfuly generated api parameters for {operation_markup}, count: {len(generated_api_parameters)}"
                )
                # filter generated_api_parameters if a pattern option is provided
                api_parameters_for_operation = generated_api_parameters
                if match_patterns:
                    pattern_matched_api_parameters = api_parameters_match_pattern(
                        generated_api_parameters, match_patterns
                    )
                    logger.debug(
                        f"Matching patterns: {match_patterns}. Filtered the generated api parameters [{len(pattern_matched_api_parameters)}/{len(generated_api_parameters)}]"
                    )
                    api_parameters_for_operation = pattern_matched_api_parameters
                for api_parameter in api_parameters_for_operation:
                    # for each parameter generated, call the actual operation
                    self.call_operation(resource_node, operation_name, api_parameter, follow_pagination=follow_pagination)
            # after calling the same operation for the different parameters
            # get all the response data made for this operation_name
            logger.debug(f"[underline][bold]Done Reading[/] {operation_markup}[/]")
            return self.search_operation_data(resource_node_name, operation_name)

        # OPERATION HAVE RELATIONS
        all_related_operations_data = {}

        # for each relation, fetch the related resource's data.
        for rel in relations_of_operation:
            rel_operation_data = self.read_operation(
                rel.resource_node_name, rel.operation_name, refresh=refresh, follow_pagination=follow_pagination
            )
            if not rel_operation_data:
                logger.debug(
                    f"[red]Failed to read related operation[/]: {rel.resource_node_name}.{rel.operation_name}"
                )
                inform_about_develeoping_custom_resource_nodes()
                return False

            # gather all their related operations data, put it under a dict
            all_related_operations_data.update({rel.operation_name: rel_operation_data})

        # send the operations_data to resource_node to create valid_api_parameters
        (
            generated_api_parameters,
            generation_error,
        ) = resource_node.generate_api_parameters_from_operation_data(
            operation_name, relations_of_operation, all_related_operations_data
        )
        
        if generated_api_parameters == [] and generation_error is None:
            # no errors, just no data available to generate api parameters
            logger.debug(
                f"[bold yellow]WARNING: There's no related resource data in your account to generate the api params for: {operation_markup}."
            )
        
        if generation_error is not None:
            logger.debug(
                f"Failed to generate api parameters for {operation_markup}: {generation_error}"
            )
            inform_about_develeoping_custom_resource_nodes()

            return False
        elif isinstance(generated_api_parameters, Iterable):
            logger.debug(
                f"Successfuly generated api parameters for [bold blue]{operation_name}[/], count: {len(generated_api_parameters)}"
            )
            # filter generated_api_parameters if a pattern option is provided
            api_parameters_for_operation = generated_api_parameters
            if match_patterns:
                pattern_matched_api_parameters = api_parameters_match_pattern(
                    generated_api_parameters, match_patterns
                )
                logger.debug(
                    f"Matching given patterns: {match_patterns}. Filtered the generated api parameters [{len(pattern_matched_api_parameters)}/{len(generated_api_parameters)}]"
                )
                api_parameters_for_operation = pattern_matched_api_parameters
            for api_parameter in api_parameters_for_operation:
                # for each parameter generated, call the actual operation
                self.call_operation(resource_node, operation_name, api_parameter, follow_pagination=follow_pagination)
        else:
            logger.debug(f"Failed to generate api parameters for {operation_markup}")

        # after calling the same operation for the different parameters
        # get all the response data made for this operation_name
        logger.debug(f"[underline][bold]Done Reading[/] {operation_markup}[/]")

        return self.search_operation_data(resource_node_name, operation_name)

    def read_resource_node(
        self,
        resource_node_name: str,
        match_patterns: Optional[List[str]] = None,
        refresh: Optional[bool] = False,
        follow_pagination: Optional[bool] = False,
    ) -> Union[Dict, bool]:
        """Reads all available operations in the given a resource node

        Args:
            resource_node_name (str): _description_
            match_patterns (List[str], optional): UNIX style patterns to filter the generated parameters. Defaults to None.
            refresh (Optional[bool], optional): Use the cached data or always make new calls. Defaults to False.
            follow_pagination (Optional[bool], optional): Follow pagination if the output is truncated.. Defaults to False.

        Returns:
            Union[Dict, bool]: Data read if successful, or False.
        """

        resource_node = self.service_node.get_resource_node_by_name(resource_node_name)
        if not resource_node:
            return False

        for operation_name in resource_node.operation_names:
            self.read_operation(
                resource_node_name, operation_name, match_patterns, refresh=refresh, follow_pagination=follow_pagination
            )
        return self.search_resource_node_data(resource_node.name)



import re
import textwrap
from typing import Dict, List, Tuple, Union
import jmespath
from terraform_import.models import TerraformImportConfig
from terraform_import.parsers import (
    parse_custom_terraform_import_configs_from_files,
    _TERRAFORM_TYPES_KEY,
)
from config import get_logger
from aws import BalconyAWS
from jinja2 import Environment

logger = get_logger(__name__)
# cache for read & parsed custom terraform import configuration .yml files
_terraform_import_configurations = None


def get_custom_terraform_import_config_dict() -> Dict:
    global _terraform_import_configurations
    if _terraform_import_configurations:
        return _terraform_import_configurations

    _terraform_import_configurations = (
        parse_custom_terraform_import_configs_from_files()
    )
    return _terraform_import_configurations


def extract_resource_tags_as_kwargs(data: dict) -> dict:
    tags_as_kwargs = {}
    tag_list = data.get("Tags", [])
    if not tag_list or not isinstance(tag_list, list):
        return tags_as_kwargs

    for tag in tag_list:
        key = tag.get("Key")
        value = tag.get("Value")
        tags_as_kwargs[f"tag_{key}"] = value
    return tags_as_kwargs


def render_jinja2_template_with_data(data, jinja2_template_str) -> List[str]:
    template = Environment().from_string(jinja2_template_str)
    result = []
    # if the data is a dict, add the key-value pairs as kwargs
    if isinstance(data, dict):
        kwargs = data.copy()
        kwargs["data"] = data
        # if there's add it as tag_Name variable
        tags_as_kwargs = extract_resource_tags_as_kwargs(data)
        kwargs.update(tags_as_kwargs)
        rendered_output = template.render(**kwargs).strip()
        result.append(rendered_output)
    # if the data is a list, render the template for each item
    elif isinstance(data, list):
        kwargs = {}
        for an_item in data:
            kwargs["data"] = an_item
            kwargs["item"] = an_item
            rendered_output = template.render(**kwargs).strip()
            rendered_list = [_.strip() for _ in rendered_output.split(" ") if _]
            result.extend(rendered_list)
    elif isinstance(data, str):
        kwargs = {
            "item": data,
            "data": data,
        }
        rendered_output = template.render(**kwargs).strip()
        result.append(rendered_output)
    return result


def gen_resource_name_and_import_id_from_op_data_(
    operation_data,
    jmespath_query,
    to_resource_name_tpl,
    id_generator_tpl,
    multiline_output=False,
):
    result = []

    resource_data = operation_data

    if not multiline_output:
        # NON MULTILINE OUTPUT
        # filter the operation data if jmespath_query is given
        # and render them one by one
        if jmespath_query:
            resource_data = jmespath.search(jmespath_query, operation_data)
            logger.debug(
                f"Filtered data using jmespath query: [bold]{jmespath_query}[/]"
            )

        for a_resource_data in resource_data:
            resource_name_list = render_jinja2_template_with_data(
                a_resource_data, to_resource_name_tpl
            )
            import_id_list = render_jinja2_template_with_data(
                a_resource_data, id_generator_tpl
            )
            result.extend(list(zip(resource_name_list, import_id_list)))
        return result
    else:
        # MULTILINE OUTPUT
        # no jmespath query given, use the whole operation data
        # assumes there's multiple lines of output from the template

        if jmespath_query:
            resource_data = jmespath.search(jmespath_query, operation_data)
            logger.debug(
                f"Filtered data using jmespath query: [bold]{jmespath_query}[/]"
            )

        multiline_resource_name_list = render_jinja2_template_with_data(
            resource_data, to_resource_name_tpl
        )
        multiline_import_id_list = render_jinja2_template_with_data(
            resource_data, id_generator_tpl
        )

        if len(multiline_resource_name_list) != len(multiline_import_id_list):
            logger.debug(
                f"[red bold]Error: multiline output lists are not equal length. {multiline_resource_name_list=} -- {multiline_import_id_list=}"
            )
            return []

        r = list(zip(multiline_resource_name_list, multiline_import_id_list))
        return r


def generate_terraform_import_block(to_resource_type, to_resource_name, import_id):
    jinja_tmpl = textwrap.dedent(
        """
    import {
        to = {{ to_resource_type }}.{{ to_resource_name }}
        id = "{{ import_id }}"
    }
    """
    )

    jinja_env = Environment()
    template = jinja_env.from_string(jinja_tmpl)
    rendered_output = template.render(
        to_resource_type=to_resource_type,
        import_id=import_id,
        to_resource_name=to_resource_name,
    ).strip()
    return rendered_output


def get_importable_resources() -> List[Tuple[str, str]]:
    custom_tf_config_dict = get_custom_terraform_import_config_dict()

    importable_services_and_resources = []
    for service_name, resource_config_dict in custom_tf_config_dict.items():
        if service_name == _TERRAFORM_TYPES_KEY:
            continue  # skip the _terraform_types cache key
        for resource_node_name, resource_config_list in resource_config_dict.items():
            for resource_config in resource_config_list:
                terraform_resource_type = resource_config.to_resource_type
                importable_services_and_resources.append(
                    (terraform_resource_type, service_name, resource_node_name)
                )
    return list(sorted(importable_services_and_resources, key=lambda x: x[1] + x[2]))


def sanitize_resource_name_and_import_ids(list_of_tuples):
    result = []
    for resource_name, import_id in list_of_tuples:
        # change anything that's not a letter, number, dash or underscore to underscore
        sanitized_resource_name = re.sub(r"[^A-Za-z0-9\-_]", "_", resource_name)
        # allow / in the import id
        # sanitized_import_id = re.sub(r"[^A-Za-z0-9\-_/]", "_", import_id)
        if sanitized_resource_name and import_id:
            result.append((sanitized_resource_name, import_id))
    return result


def get_import_config_for(
    service: str = None,
    resource_node: str = None,
    terraform_resource_type: str = None,
) -> List[TerraformImportConfig]:
    custom_tf_config_dict = get_custom_terraform_import_config_dict()

    if service and resource_node:
        config_for_resource_node = custom_tf_config_dict.get(service, {}).get(
            resource_node, []
        )
        return config_for_resource_node
    elif terraform_resource_type:
        config_list_for_tf_type = custom_tf_config_dict.get(
            _TERRAFORM_TYPES_KEY, {}
        ).get(terraform_resource_type, False)
        if not config_list_for_tf_type:
            return False
        return config_list_for_tf_type
    return False


def generate_import_block_from_import_config(
    balcony_client: BalconyAWS,
    tf_import_config: TerraformImportConfig,
    follow_pagination: bool = False,
) -> List[str]:
    tf_import_blocks: List[str] = []

    operation_name = tf_import_config.operation_name
    service = tf_import_config.service
    resource_node = tf_import_config.resource_node
    jmespath_query = tf_import_config.jmespath_query

    if not tf_import_config:
        logger.debug(
            f"[red bold]No custom terraform import config found for {service}.{resource_node}. Please check out docs https://oguzhan-yilmaz.github.io/balcony/ for more info on developing it your own."
        )
        return False

    # read the operation
    operation_data = balcony_client.read_operation(
        service_name=service,
        resource_node_name=resource_node,
        operation_name=operation_name,
        follow_pagination=follow_pagination,
    )

    if not operation_data:
        logger.debug(f"No data found for {service}.{resource_node}.{operation_name}.")
        return False

    resource_name_and_import_ids = gen_resource_name_and_import_id_from_op_data_(
        operation_data,
        jmespath_query,
        tf_import_config.to_resource_name_jinja2_template,
        tf_import_config.id_generator_jinja2_template,
        multiline_output=tf_import_config.multiline_output,
    )

    # Replace unsupported chars from the to-resource-name with underscore
    sanitized_resource_name_and_import_ids = sanitize_resource_name_and_import_ids(
        resource_name_and_import_ids
    )

    logger.debug(
        f"Found {len(sanitized_resource_name_and_import_ids)} resources to import."
    )
    for a_tuple in sanitized_resource_name_and_import_ids:
        to_resource_name, import_id = a_tuple
        tf_import_block = generate_terraform_import_block(
            tf_import_config.to_resource_type, to_resource_name, import_id
        )
        tf_import_blocks.append(tf_import_block)
    return tf_import_blocks


def generate_import_block_for_resource(
    balcony_client: BalconyAWS,
    service: str = None,
    resource_node: str = None,
    terraform_resource_type: str = None,
    follow_pagination: bool = False,
):
    operation_markup = f"[bold][green]{service}[/].[blue]{resource_node}[/][/]"
    if (not service or not resource_node) and terraform_resource_type:
        operation_markup = f"[bold cyan]{terraform_resource_type}[/]"

    logger.debug(f"[underline bold][green]Starting to Generate[/] Terraform import blocks for[/] {operation_markup}.")
    resulting_tf_import_blocks = []
    tf_import_configs = get_import_config_for(
        service, resource_node, terraform_resource_type
    )
    if not tf_import_configs:
        logger.debug(
            f"[red bold]No custom terraform import config found for {service}.{resource_node}. Please check out docs https://oguzhan-yilmaz.github.io/balcony/ for more info on developing it your own."
        )
        return False

    for tf_import_config in tf_import_configs:
        tf_import_blocks = generate_import_block_from_import_config(
            balcony_client, tf_import_config, follow_pagination
        )
        if not tf_import_blocks:
            logger.debug(
                f"[red bold]No data found for {tf_import_config.to_resource_type} — {tf_import_config.resource_node}.{tf_import_config.operation_name}."
            )

        if tf_import_blocks and isinstance(tf_import_blocks, list):
            resulting_tf_import_blocks.extend(tf_import_blocks)
    logger.debug(f"[underline bold][green]Done Generating[/] Terraform import blocks for[/] {operation_markup}.")
    return resulting_tf_import_blocks



from aws import BalconyAWS
from terraform_import.main import main as terraform_import_main

balcony_aws = BalconyAWS()
tf_import_blocks = terraform_import_main(balcony_aws, "ec2", "Instances")
print('\n\n'.join(tf_import_blocks))

tf_import_blocks = terraform_import_main(balcony_aws, "s3", "Buckets")
print('\n\n'.join(tf_import_blocks))
tf_import_blocks


# from utils import is_word_in_a_list_of_words
# from typing import Dict, List, Union
# from nodes import PAGINATION_TOKEN_KEYS

# def get_pagination_token_output_to_parameter_name_mapping(
#     self, operation_name: str
# ) -> Union[Dict[str, str], bool]:
#     """Some Operations paginate their output using Pagination Token Parameters defined in `PAGINATION_TOKEN_KEYS`.

#     Args:
#         resource_node (ResourceNode): _description_
#         operation_name (str): Name of the operation in the resource_node.

#     Returns:
#         Union[Dict[str, str], bool]: False or A dictionary with `parameter_name` and `output_key` keys. e.g. `{"parameter_name":"nextToken", "output_key": "NextToken"}`.
#     """
#     resource_node = self

#     # find all parameter names that are in `PAGINATION_TOKEN_KEYS`, case insensitive
#     parameter_names = [
#         pn
#         for pn in resource_node.get_all_parameter_names_from_operation_name(
#             operation_name
#         )
#         if is_word_in_a_list_of_words(pn, PAGINATION_TOKEN_KEYS)
#     ]

#     #  find all output keys that are in `PAGINATION_TOKEN_KEYS`, case insensitive
#     output_keys = [
#         ok
#         for ok in resource_node.get_all_output_keys_from_operation_name(
#             operation_name
#         )
#         if is_word_in_a_list_of_words(ok, PAGINATION_TOKEN_KEYS)
#     ]

#     if output_keys and parameter_names

#         return {
#             "parameter_name": parameter_names[0],
#             "output_key": output_keys[0],
#         }
#     return False


# def main():
#     baws = BalconyAWS()
#     aws_services = baws.get_available_service_names()
#     for aws_service_name in aws_services[aws_services.index("lexv2-runtime") :]:
#         print(aws_service_name)
#         service_node = baws.get_service_node(aws_service_name)
#         for resource_node in service_node.get_resource_nodes():
#             for operation_name in resource_node.get_operation_names():
#                 a = get_pagination_token_output_to_parameter_name_mapping(
#                     resource_node, operation_name
#                 )
#                 print(a)

#                 # "nexttoken, continuationtoken, marker"
#                 # rpp = [_ for _ in rp if "token" in _.lower() or "marker" in _.lower()]
#                 # opp = [_ for _ in op if "token" in _.lower() or "marker" in _.lower()]
#                 # if rpp or opp:
#                 #     print(
#                 #         f"  {operation_name} :: {','.join(rpp)} --::-- {','.join(opp)}"
#                 #     )


# def __():
#     pass


# main()



from utils import find_all_yaml_files
from config import (
    get_logger,
    YAML_TF_IMPORT_CONFIGS_DIRECTORY,
    USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY,
)
from terraform_import.models import (
    CustomTerraformImportConfigFile,
    # TerraformImportConfig,
)
from typing import Union, Tuple
import yaml
from collections import defaultdict

logger = get_logger(__name__)
_TERRAFORM_TYPES_KEY = "_terraform_types"


def parse_json_to_tf_import_config(
    input_configuration_dict: str,
) -> Union[bool, CustomTerraformImportConfigFile]:
    try:
        return CustomTerraformImportConfigFile(**input_configuration_dict)
    except Exception as e:
        logger.debug(f"Error while parsing user defined tf import config: {str(e)}")
    return False


def parse_yaml_file_to_tf_import_config(
    yaml_file_path: str,
) -> Tuple[CustomTerraformImportConfigFile, Union[Exception, None]]:
    """Parses given YAML file and returns a pydantic model: YamlService object

    Args:
        yaml_file_path (str): Filepath for the yaml file

    Returns:
        YamlService: Input data parsed as a pydantic YamlService object
        Exception or `None`: Exception if raised during parsing, or None if successful
    """
    try:
        with open(yaml_file_path, "r") as f:
            input_data = yaml.load(f, Loader=yaml.FullLoader)
        return CustomTerraformImportConfigFile(**input_data), None
    except Exception as e:
        return False, e


def parse_custom_terraform_import_configs_from_files():
    terraform_configurations_dict = defaultdict(lambda: defaultdict(dict))

    parse_directories = [
        YAML_TF_IMPORT_CONFIGS_DIRECTORY,
    ]  # noqa
    _are_same_directories = (
        USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY
        == YAML_TF_IMPORT_CONFIGS_DIRECTORY.as_posix()
    )
    if USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY and not _are_same_directories:
        # might not be defined
        parse_directories.append(USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY)

    for yaml_directory in parse_directories:
        config_filenames = find_all_yaml_files(yaml_directory)
        logger.debug(
            f"Terraform Import Configuration Registry: Found {len(config_filenames)} yaml files in {yaml_directory}. Starting to parse & validate them."
        )
        # parse them
        for conf_filename in config_filenames:
            conf, error = parse_yaml_file_to_tf_import_config(conf_filename)
            if error:
                logger.error(f"Error while parsing {conf_filename}: {error}")
                continue
            for tf_config in conf.import_configurations:
                # doing it this way allows overrides from the user defined configs
                r_node = terraform_configurations_dict.get(tf_config.service, {}).get(
                    tf_config.resource_node, False
                )
                if r_node is False:
                    terraform_configurations_dict[tf_config.service][
                        tf_config.resource_node
                    ] = [tf_config]
                else:
                    terraform_configurations_dict[tf_config.service][
                        tf_config.resource_node
                    ].append(tf_config)
                # add it to _TERRAFORM_TYPES_KEY
                tf_conf_list = terraform_configurations_dict[_TERRAFORM_TYPES_KEY].get(
                    tf_config.to_resource_type, []
                )
                if not tf_conf_list:
                    terraform_configurations_dict[_TERRAFORM_TYPES_KEY][
                        tf_config.to_resource_type
                    ] = [tf_config]
                else:
                    terraform_configurations_dict[_TERRAFORM_TYPES_KEY][
                        tf_config.to_resource_type
                    ].append(tf_config)
    return terraform_configurations_dict



from config import get_logger
from typing import List, Dict
from aws import BalconyAWS
from nodes import ServiceNode
logger = get_logger(__name__)

def export_boto3_operations_by_service(balcony_aws: BalconyAWS) -> List[Dict[str, List[str]]]:
    service_and_operation_names = {}
    for service_name in balcony_aws.get_available_service_names():
        logger.debug(f"Getting boto3 operations for service: {service_name}")
        service_node: ServiceNode = balcony_aws.get_service_node(service_name)
        operation_names = service_node.get_operation_names()
        service_and_operation_names[service_name] = operation_names
    return service_and_operation_names



# from ..nodes import ResourceNode
# from ..config import get_logger

# logger = get_logger(__name__)



# class Cluster(ResourceNode, service_name="ecs", name="Clusters"):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
    
#     def generate_jmespath_selector_from_relations(self, operation_name, relation_list):
#         r = super().generate_jmespath_selector_from_relations(operation_name, relation_list)
#         if operation_name == 'ListClusters':
#             return '[*].{cluster: clusterArns}[][]'
#         return r
    
#     # def define_extra_relations(self):
#     #     r= super().define_extra_relations()
#     #     return [{
#     #         "service_name": "ecs",
#     #         "resource_node_name": "Clusters",
#     #         "required_shape_name": "cluster",
#     #         "target_shape_name": "cluster",
#     #         "target_shape_type": "string",
#     #         "operation_name": "ListClusters",
#     #         "target_path": "clusterArns"
#     #     }]
        

# class Tasks(ResourceNode, service_name="ecs", name="Tasks"):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)

#     def get_required_parameter_names_from_operation_name(self, operation_name):
#         r = super().get_required_parameter_names_from_operation_name(operation_name)
#         if operation_name == 'DescribeTasks':
#             return ['tasks']
#         if operation_name == 'ListTasks':
#             return ['cluster']
#         return r
    
#     # def define_extra_relations(self):
#     #     return [
#     #         {
#     #             "service_name": "ecs",
#     #             "resource_node_name": "Tasks",
#     #             "required_shape_name": "tasks",
#     #             "target_shape_name": "taskArns",
#     #             "target_shape_type": "string",
#     #             "operation_name": "ListTasks",
#     #             "target_path": "taskArns"
#     #         }
#     #     ]
     
# class Services(ResourceNode, service_name="ecs", name="Services"):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)

#     def get_required_parameter_names_from_operation_name(self, operation_name):
#         r = super().get_required_parameter_names_from_operation_name(operation_name)
#         if operation_name == 'ListServices':
#             return ['cluster']
#         return r
    
#     def generate_jmespath_selector_from_relations(self, operation_name, relation_list) -> str:
#         if operation_name == 'ListServices':
#             return "[*].clusterArns[].{cluster: @}"
#         if operation_name == 'DescribeServices':
#             return "[*].serviceArns[].{services: [@]}"
        
#         return super().generate_jmespath_selector_from_relations(operation_name, relation_list)

    
#     def define_extra_relations(self):
#         return [
#             {
#                 "service_name": "ecs",
#                 "resource_node_name": "Services",
#                 "required_shape_name": "services",
#                 "target_shape_name": "serviceArns",
#                 "target_shape_type": "string",
#                 "operation_name": "ListServices",
#                 "target_path": "[*].serviceArns"
#             },
#             {
#                 "service_name": "ecs",
#                 "resource_node_name": "Clusters",
#                 "required_shape_name": "cluster",
#                 "target_shape_name": "cluster",
#                 "target_shape_type": "string",
#                 "operation_name": "ListClusters",
#                 "target_path": "[*].clusterArns"
#             }
#         ]

#     # def get_operations_relations(self, operation_name):
#     #     r = super().get_operations_relations(operation_name)
#     #     if operation_name == 'DescribeServices':
#     #         return [
#     #         #     {
#     #         #     "service_name": "ecs",
#     #         #     "resource_node_name": "Services",
#     #         #     "required_shape_name": "services",
#     #         #     "target_shape_name": "serviceArns",
#     #         #     "target_shape_type": "string",
#     #         #     "operation_name": "ListServices",
#     #         #     "target_path": "serviceArns"
#     #         # },
#     #         {
#     #             "service_name": "ecs",
#     #             "resource_node_name": "Services",
#     #             "required_shape_name": "cluster",
#     #             "target_shape_name": "cluster",
#     #             "target_shape_type": "string",
#     #             "operation_name": "ListClusters",
#     #             "target_path": "[*].clusterArns"
#     #         }
#     #         ], None
#     #     return r
    
#     # def define_extra_relations(self):
#     #     r= super().define_extra_relations()
   


from nodes import ResourceNode
from config import get_logger
from typing import List, Dict, Tuple

logger = get_logger(__name__)


class BuildsForProject(ResourceNode, service_name="codebuild", name="BuildsForProject"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_operations_relations(self, operation_name) -> Tuple[List[Dict], None]:
        result, err = super().get_operations_relations(operation_name)
        # manipulate `result` or return your own value
        return result, err



from typing import List, Optional
from pydantic import BaseModel  # , validator


class TerraformImportConfig(BaseModel):
    service: str  # ec2
    resource_node: str  # Instances
    operation_name: str  # DescribeInstances
    jmespath_query: Optional[str]  # Reservations[].Instances[]
    to_resource_type: str  # aws_instance
    to_resource_name_jinja2_template: str  # "{{ InstanceId }}"
    id_generator_jinja2_template: str
    multiline_output: Optional[bool]

    # TODO: code the validators
    # @validator("to_resource_name_jinja2_template")
    # def validate_to_resource_name_jinja2_template(cls, value):
    #     # Add your validation logic here
    #     # You can raise a ValueError if the value is invalid
    #     return value

    # @validator("id_generator_jinja2_template")
    # def validate_id_generator_jinja2_template(cls, value):
    #     # Add your validation logic here
    #     # You can raise a ValueError if the value is invalid
    #     return value


class MaintainersBlock(BaseModel):
    """List of maintainers for the terraform import config file."""

    name: str
    github: Optional[str]
    email: Optional[str]


class CustomTerraformImportConfigFile(BaseModel):
    maintainers: Optional[List[MaintainersBlock]]
    import_configurations: List[TerraformImportConfig]



# from ..nodes import ResourceNode
# from ..config import get_logger

# logger = get_logger(__name__)

# class Function(ResourceNode, service_name="lambda", name="Function"):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
    
#     def define_extra_relations(self):
#         r= super().define_extra_relations()
#         return [{
#             "service_name": "lambda",
#             "resource_node_name": "Function",
#             "operation_name": "ListFunctions",
#             "required_shape_name": "FunctionName",
#             "target_shape_name": "FunctionName",
#             "target_shape_type": "string",
#             "target_path": "Functions[*].FunctionName"
#         }]


import os
import random
import string
import textwrap
from typing import List, Union
import jmespath
import yaml
from terraform_import.importer import render_jinja2_template_with_data
from terraform_import.parsers import parse_json_to_tf_import_config
from config import (
    get_logger,
    get_rich_console,
    USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY,
)
from aws import BalconyAWS
from rich.prompt import Prompt, Confirm
import typer
from botocore.utils import ArgumentGenerator
from rich.panel import Panel
from rich.padding import Padding
from rich.pretty import Pretty
from rich.syntax import Syntax

logger = get_logger(__name__)
console = get_rich_console()


def get_operation_names_of_a_resource(
    balcony_aws: BalconyAWS, service: str, resource_name: str
):
    service_node = balcony_aws.get_service_node(service)
    if not service_node:
        return False
    resource_node = service_node.get_resource_node_by_name(resource_name)
    if not resource_node:
        return False
    operation_names = resource_node.operation_names
    return operation_names


def get_documentation_panel_for_operation(
    balcony_aws: BalconyAWS, service: str, resource_name: str, operation_name: str
):
    service_node = balcony_aws.get_service_node(service)
    if not service_node:
        return False
    resource_node = service_node.get_resource_node_by_name(resource_name)
    if not resource_node:
        return False
    operation_docs_panel = resource_node._rich_operation_details_panel(
        operation_name, remove_input_shape=True, remove_documentation=True
    )
    return operation_docs_panel


def get_operation_model_for_operation(
    balcony_aws: BalconyAWS, service: str, resource_name: str, operation_name: str
):
    service_node = balcony_aws.get_service_node(service)
    if not service_node:
        return False
    resource_node = service_node.get_resource_node_by_name(resource_name)
    if not resource_node:
        return False
    operation_model = resource_node.get_operation_model(operation_name)
    return operation_model


def gen_mock_output_data_for_operation(
    balcony_aws, service, resource_name, operation_name
):
    arg_gen = ArgumentGenerator(use_member_names=True)
    op_model = get_operation_model_for_operation(
        balcony_aws, service, resource_name, operation_name
    )
    mock_output_data = arg_gen.generate_skeleton(op_model.output_shape)

    return mock_output_data


def gen_balcony_tf_import_config_yaml(user_def_import_configuration_list: List[dict]):
    yaml_dict = {
        "maintainers": [{"name": "Your Name"}],
        "import_configurations": user_def_import_configuration_list,
    }

    parsed = parse_json_to_tf_import_config(yaml_dict)
    if not parsed:
        return False

    yaml_output = yaml.dump(yaml_dict)
    return yaml_output


def generate_google_search_uri(keywords):
    search_terms = "+".join(keywords)
    return f"https://www.google.com/search?q={search_terms}"


def save_tf_input_config_to_user_defined_yaml_dir(filename: str, yaml_content: str):
    if not USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY:
        return False

    filepath = os.path.join(USER_DEFINED_YAML_TF_IMPORT_CONFIGS_DIRECTORY, filename)

    with open(filepath, "w") as f:
        f.write(yaml_content)

    return filepath


def interactive_help(balcony_aws: BalconyAWS, service: str, resource_name: str):
    console.print(
        Padding(
            Panel(  # noqa
                textwrap.dedent(
                    f"""
                Welcome to the Balcony interactive help for generating Terraform import configurations for AWS Services.
                
                You'll be asked a series of questions to help you generate the configuration.                        
                """
                ),  # noqa
                title="[bold blue]Balcony Interactive Help - Generate Terraform Import Configurations",
            ),
            (1, 2),
        ),
    )

    operation_names = get_operation_names_of_a_resource(
        balcony_aws, service, resource_name
    )
    if not operation_names:
        console.print(
            f"[red bold]No operations found for {service}.{resource_name}[/red bold]"
        )
        raise typer.Exit(code=-1)

    console.print()
    console.rule("[bold cyan underline]Operation Selection")
    console.print()
    console.print("balcony may have multiple Operations for a Resource.")
    console.print(
        "Select which operation to use for generating the import configuration."
    )
    console.print(
        "Run the same command with '-ls, --list --screen' option added to see all operations of a Resource."
    )
    console.print()
    selected_operation_name = Prompt.ask(
        "[green bold] Select which operation",
        choices=operation_names,
        default=operation_names[0],
        console=console,
    )

    operation_doc_panel = get_documentation_panel_for_operation(
        balcony_aws, service, resource_name, selected_operation_name
    )
    # with console.pager(styles=True):
    #     console.print(operation_doc_panel)
    console.print(operation_doc_panel)
    console.print()

    boto3_google_search_uri = generate_google_search_uri(
        ["terraform", "provider", "aws", service, resource_name]
    )
    terraform_aws_docs_google_search_uri = generate_google_search_uri(
        ["boto3", service, resource_name, selected_operation_name]
    )

    console.print(
        Padding(
            Panel(  # noqa
                textwrap.dedent(
                    f"""
                Google — boto3 docs: [bold cyan]{terraform_aws_docs_google_search_uri}[/]
                
                Google — Terraform docs: [bold cyan]{boto3_google_search_uri}[/]                     
                """,
                ),  # noqa
                title="[bold blue]Google Search Links",
            ),
            (0, 2),
        ),
    )
    console.print()

    # behave like balcony does, as it reads lists of pages, so has a list of (responses,)
    mock_output_data_list = [
        gen_mock_output_data_for_operation(
            balcony_aws, service, resource_name, selected_operation_name
        ),
        gen_mock_output_data_for_operation(
            balcony_aws, service, resource_name, selected_operation_name
        ),
        # gen_mock_output_data_for_operation(
        #     balcony_aws, service, resource_name, selected_operation_name
        # ),
    ]

    console.rule("[bold cyan underline]JMESPath Query")
    console.print()
    confirm_jmespath_selector = False
    jmespath_selector = None
    jmespath_filtered_data = None
    do_ask_for_jmespath_selector = True
    while do_ask_for_jmespath_selector:
        input_jmespath_selector = Prompt.ask(
            "[green bold] Enter a jmespath query to filter the mock output data",
            console=console,
        )

        if not input_jmespath_selector:
            console.print("[red bold]You must enter a jmespath query.[/red bold]")
            continue

        try:
            _filtered_data = jmespath.search(
                input_jmespath_selector, mock_output_data_list
            )
        except jmespath.exceptions.ParseError as e:
            _filtered_data = False

        if not _filtered_data:
            console.print(
                f"[red bold]No data found for jmespath query {input_jmespath_selector}. Try again.[/red bold]"
            )
            continue  # loop again

        def _easy_on_the_eyes_dict(list_of_dicts):
            for _dict in list_of_dicts:
                if isinstance(_dict, dict):
                    for key, value in _dict.items():
                        if isinstance(value, dict):
                            _dict[key] = "{...omitted...}"
                        elif isinstance(value, list):
                            _dict[key] = "[...omitted...]"
            return list_of_dicts

        console.print_json(data=_easy_on_the_eyes_dict(_filtered_data), default=str)

        confirm_jmespath_selector = Confirm.ask(
            f"[yellow bold]Does the above data look correct? Are you sure to use [bold blue]{input_jmespath_selector}[/] jmespath query?[/]",
            console=console,
        )

        # check for confirmation and existence of actually filtered data
        if confirm_jmespath_selector and _filtered_data:
            jmespath_selector = input_jmespath_selector
            jmespath_filtered_data = _filtered_data
            # break out of the loop
            do_ask_for_jmespath_selector = False

            filtered_data_type_is_dict = isinstance(jmespath_filtered_data[0], dict)
            if not filtered_data_type_is_dict:
                # warn about the usage
                console.print(
                    Padding(
                        Panel(
                            f"[blue bold] Note:[/][blue] You've selected a non-dict data. You can access the elements with 'item' keyword.",
                            expand=False
                        ),
                        (1, 1),
                    )
                )
            continue

    # randomize the data ids, etc
    for filtered_mock_data in jmespath_filtered_data:
        random_string = "".join(random.choices(string.ascii_letters, k=4))
        if isinstance(filtered_mock_data, dict):
            for key, value in filtered_mock_data.items():
                if isinstance(value, str):
                    filtered_mock_data[key] = f"{value}-{random_string}"
        if isinstance(filtered_mock_data, list):
            filtered_mock_data = [
                f"{_m_data}-{random_string}" if isinstance(_m_data, str) else _m_data
                for _m_data in filtered_mock_data
            ]
        if isinstance(filtered_mock_data, str):
            filtered_mock_data = f"{filtered_mock_data}-{random_string}"
    console.print()
    console.rule("[cyan underline]Jinja2 Template for [bold]Terraform to Resource Name")
    console.print()
    console.print(
        "[yellow bold] The j2 template you provide must select unique values, or terraform will fail.[/]"
    )
    console.print(
        "You can directly access the attrs. by their name (e.g. PolicyId, RoleArn, ClusterName etc.)"
    )
    console.print(
        "You can also access the resource tags by their key prefixed with `tag_` (e.g. `tag_Name`, `tag_Environment` etc.)"
    )
    console.print(
        "Use `or` keyword to select the first non-empty value (e.g. `tag_Name or tag_Billing or InstanceId or PolicyId`)"
    )
    console.print()

    do_ask_for_j2_tpl_resource_name = True
    while do_ask_for_j2_tpl_resource_name:
        input_j2_tpl_resource_name = Prompt.ask(
            "[green bold] Enter a jinja2 template to generate the to_resource_name. Your input will be wrapped with '{{ }}'",
            console=console,
        )
        # wrap the input with {{ }}
        j2_tpl_resource_name = f"{{{{ {input_j2_tpl_resource_name} }}}}"

        # render the jinja2 template with the filtered data
        to_resource_name_list = []
        for filtered_datum in jmespath_filtered_data:
            _rendered_data_list = render_jinja2_template_with_data(
                filtered_datum, j2_tpl_resource_name
            )
            to_resource_name_list.extend(_rendered_data_list)

        console.print(
            Panel(
                Pretty(to_resource_name_list),
                title=f"Rendered your j2 template: {j2_tpl_resource_name}",
            )
        )
        console.print()
        confirm_j2_tpl_resource_name = Confirm.ask(
            f"[yellow bold]Are you sure to use [bold blue]{j2_tpl_resource_name}[/] ?[/]",
            console=console,
        )

        if confirm_j2_tpl_resource_name and to_resource_name_list:
            do_ask_for_j2_tpl_resource_name = False
            continue

    console.print()
    console.rule("[cyan underline]Jinja2 Template for [bold]Import ID")
    console.print()

    do_ask_for_j2_tpl_import_id = True
    while do_ask_for_j2_tpl_import_id:
        input_j2_tpl_import_id = Prompt.ask(
            "[green bold] Enter a jinja2 template to generate the Import ID. Your input will be wrapped with '{{ }}'",
            console=console,
        )
        # wrap the input with {{ }}
        j2_tpl_import_id = f"{{{{ {input_j2_tpl_import_id} }}}}"

        # render the jinja2 template with the filtered data
        import_id_list = []
        for filtered_datum in jmespath_filtered_data:
            _rendered_data_list = render_jinja2_template_with_data(
                filtered_datum, j2_tpl_import_id
            )
            import_id_list.extend(_rendered_data_list)

        console.print(
            Panel(
                Pretty(import_id_list),
                title=f"Rendered your j2 template: {j2_tpl_import_id}",
            )
        )
        console.print()
        confirm_j2_tpl_import_id = Confirm.ask(
            f"[yellow bold]Are you sure to use [bold blue]{j2_tpl_import_id}[/] ?[/]",
            console=console,
        )

        if confirm_j2_tpl_import_id and import_id_list:
            do_ask_for_j2_tpl_import_id = False
            continue

    console.print()
    console.rule("[bold cyan underline]Terraform Resource Type")
    console.print()

    do_ask_for_tf_resource_type = True
    while do_ask_for_tf_resource_type:
        to_terraform_resource_type = Prompt.ask(
            "[green bold]Enter Terraform Resource Type (e.g. aws_instance)",
            default=False,
            console=console,
        )

        confirm_tf_resource_type = Confirm.ask(
            f"[yellow bold]Are you sure to use [bold blue]{to_terraform_resource_type}[/] ?[/]",
            console=console,
        )
        if confirm_tf_resource_type:
            do_ask_for_tf_resource_type = False
            continue

    console.print()
    console.rule("")
    console.print()

    user_def_import_configurations = [
        {
            "service": service,
            "resource_node": resource_name,
            "operation_name": selected_operation_name,
            "jmespath_query": jmespath_selector,
            "to_resource_type": to_terraform_resource_type,
            "to_resource_name_jinja2_template": j2_tpl_resource_name,
            "id_generator_jinja2_template": j2_tpl_import_id,
        }
    ]

    yaml_output = gen_balcony_tf_import_config_yaml(user_def_import_configurations)
    # check yaml output validity

    if not yaml_output:
        console.print(
            f"[red bold]Error while generating yaml output. Enable --debug to see more messages.[/red bold]"
        )
        return False

    console.print("Your import-configuration yaml output:")
    console.print()
    console.print(
        Syntax(yaml_output, "yaml", theme="monokai", line_numbers=False)
    )  # noqa

    gen_filename = f"{service}-{resource_name}-wizard.yaml"
    saved_filepath = save_tf_input_config_to_user_defined_yaml_dir(
        gen_filename, yaml_output
    )

    if not saved_filepath:
        console.print(
            f"[yellow bold]Warning:[/] You can set a custom directory for your import configurations with [bold]BALCONY_TERRAFOM_IMPORT_CONFIG_DIR[/] environment variable."
        )
        console.print(
            f"If you set this, balcony will read your import configurations from that directory. And terraform-wizard will auto-save your generated import configuration to that directory."
        )
        console.print(
            Syntax(
                f"export BALCONY_TERRAFOM_IMPORT_CONFIG_DIR=~/balcony-tf-import-configs",
                "bash",
                theme="monokai",
                line_numbers=False,
            )
        )

    elif saved_filepath:
        console.print(
            f"[bold]Since you've set the env. var. 'BALCONY_TERRAFOM_IMPORT_CONFIG_DIR', auto save is enabled.[/]"
        )

        console.print(
            f"[green bold]SUCCESS: Your import-configuration yaml output saved to {saved_filepath}[/]"
        )

    return True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balcony', 'balcony.custom_nodes', 'balcony.terraform_import']

package_data = \
{'': ['*'], 'balcony': ['custom_tf_import_configs/*', 'custom_yamls/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'boto3>=1.24.80,<2.0.0',
 'inflect>=6.0.0,<7.0.0',
 'jmespath>=1.0.1,<2.0.0',
 'mkdocs-autorefs>=0.4.1,<0.5.0',
 'mkdocs-material>=8.5.7,<10.0.0',
 'mkdocstrings[python]>=0.21.2,<0.22.0',
 'pydantic>=1.10.7,<2.0.0',
 'rich>=13.3.4,<14.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['balcony = balcony.cli:run_app']}

setup_kwargs = {
    'name': 'balcony',
    'version': '0.2.2',
    'description': 'Read any resource in your AWS Account. You can generate terraform code for them, too.',
    'long_description': '# balcony\n\n\n<div style="display: flex;">\n  <a href="https://github.com/oguzhan-yilmaz/balcony/actions/workflows/docker-publish.yml"><img src="https://github.com/oguzhan-yilmaz/balcony/actions/workflows/docker-publish.yml/badge.svg" alt="Build and publish a Docker image to ghcr.io"></a>\n  <span style="width: 5px"></span>\n\n<a href="https://github.com/oguzhan-yilmaz/balcony/actions/workflows/pages/pages-build-deployment"><img src="https://github.com/oguzhan-yilmaz/balcony/actions/workflows/pages/pages-build-deployment/badge.svg" alt="Build and Deploy Documentation website"></a>\n</div>\n\n\nbalcony is a modern CLI tool that with some killer features:\n\n- Auto-fill the required parameters for AWS API calls \n- Read the JSON data of any AWS resource in your account\n- Generate [Terraform Import Blocks](https://developer.hashicorp.com/terraform/language/import)\n- Generate actual `.tf` Terraform Resource code\n\nbalcony uses _read-only_ operations, it does not take any action on the used AWS account.\n\n\n### [Visit the Documentation Website](https://oguzhan-yilmaz.github.io/balcony/quickstart/)\n<!-- ### [**Go to QuickStart Page to get started using _balcony_**](quickstart.md) -->\n\n### Installation\n\n```bash\npip3 install balcony\n```\n\nVisit [**Installation & QuickStart Page**](https://oguzhan-yilmaz.github.io/balcony/quickstart/) to get started using _balcony_\n\n```bash  title="Basic usage"\n# see options\nbalcony\n\n# list available resources of ec2\nbalcony aws ec2 \n\n# read a resource\nbalcony aws s3 Buckets\n\n# show documentation\nbalcony aws iam Policy --list\n\n# generate terraform import blocks for a resource\nbalcony terraform-import s3 Buckets\n```\n\n\n## Features\n\n### Read any AWS Resource\n\n`balcony aws <service> <resource-name> --paginate` command reads all resources of a given type in your AWS account.\n\nRelated Docs: [QuickStart](https://oguzhan-yilmaz.github.io/balcony/quickstart/)\n\n![](https://raw.githubusercontent.com/oguzhan-yilmaz/balcony-assets/main/gifs/aws-read-resource.gif)\n\n\n---\n\n### Generate Terraform Import Blocks\n\nTerraform v1.5 introduced [import blocks](https://developer.hashicorp.com/terraform/language/import) that allows users to define their imports as code.\n\n`balcony terraform-import <service> <resource-name>` command generates these import blocks for you.\n\n`balcony terraform-import --list` to see the list of supported resources.\n\nRelated Docs: [Generate Terraform Import Blocks](https://oguzhan-yilmaz.github.io/balcony/terraform-import/)\nRelated Docs: [Balcony Terraform Import Support Matrix](https://oguzhan-yilmaz.github.io/balcony/terraform-import-support-matrix/)\n\n\n\n![](https://raw.githubusercontent.com/oguzhan-yilmaz/balcony-assets/main/gifs/terraform-import-blocks-example.gif)\n\n\n---\n\n### Generate actual Terraform Resource Code \n\n\nIf you have:\n\n- initialized terraform project\n- `import_blocks.tf` file that\'s generated with `balcony terraform-import` command\n\nyou can run `terraform plan -generate-config-out=generated.tf` to generate actual `.tf` resource code.\n\nThis feature is achieved with the [balcony-terraform-import Docker Image](https://github.com/oguzhan-yilmaz/balcony/pkgs/container/balcony-terraform-import).\n\n\nRelated Docs: [Generate Terraform Code with Docker Image](https://oguzhan-yilmaz.github.io/balcony/terraform-import-docker/)\n\n![](https://raw.githubusercontent.com/oguzhan-yilmaz/balcony-assets/main/gifs/docker-gen-tf-code-ec2-insances-example.gif)\n\n\n---\n\n### Interactive Wizard to create balcony import configurations \n\nBalcony doesn\'t know how to create terraform `import blocks` for all of the AWS resources.\n\nIt can be taught how to do it by creating `import-configurations` yaml files, but it\'s a manual process. This is where the interactive wizard comes in.\n\nInteractive Wizards asks you required questions to automatically create the `import-configurations` yaml files.\n\nRelated Docs: [Terraform Import Configuration Wizard](https://oguzhan-yilmaz.github.io/balcony/terraform-import-wizard/)\n\n![](https://raw.githubusercontent.com/oguzhan-yilmaz/balcony-assets/main/gifs/terraform-wizard-security-groups-example.gif)\n',
    'author': 'Oguzhan Yilmaz',
    'author_email': 'oguzhanylmz271@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from nodes import ResourceNode
from config import get_logger
from relations import Relation
import jmespath

logger = get_logger(__name__)

class GroupPolicy(ResourceNode, service_name="sqs", name="QueueUrl"):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_operations_relations(self, operation_name: str):
        if operation_name == "GetQueueUrl":
            return [
                Relation(
                    **{
                        "service_name": "sqs",
                        "resource_node_name": "Queues",
                        "operation_name": "ListQueues",
                        "required_shape_name": "--ommitted--",
                        "target_shape_name": "--ommitted--",
                        "target_shape_type": "--ommitted--",
                        "target_path": "--ommitted--",
                    }
                )
            ], None

        return super().get_operations_relations(operation_name)

    def generate_api_parameters_from_operation_data(
        self, operation_name, relations_of_operation, related_operations_data
    ):
        if operation_name == "GetQueueUrl":
            generated_api_parameters = []

            # select non-empty PolicyNames lists
            queue_urls = jmespath.search(
                "ListQueues[].QueueUrls[]", related_operations_data
            )
            for queue_url in queue_urls:
                queue_name = queue_url.split("/")[-1]
                generated_api_parameters.append({"QueueName": queue_name})
            return generated_api_parameters, None

        # if this is NOT our selected operation, let the normal flow run
        return super().generate_api_parameters_from_operation_data(
            operation_name, relations_of_operation, related_operations_data
        )



from config import get_logger
from typing import Dict

logger = get_logger(__name__)


class Error(Exception):
    def __init__(self, message: str, context: Dict = None) -> None:
        super().__init__(message)  # Call the base class constructor
        self.message = message
        self.context = context
        # error_log = self.create_error_log()
        # if error_log:
        #     logger.debug(error_log)
